# Copyright 2024 RelationalAI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sys import stderr


def info(s: str): print("[Info]: " + s, file=stderr)


def warn(s: str): print("[Warning]: " + s, file=stderr)


# Generate a temporary variable name for use in Schema assembly
# using rules declared as defs in Relation contexts.
#
def tempvar(i: int) -> str:
    return f"_x{i}"
